Attributes deprecated(since="mathcomp 2.5.0",
  note="Use 'all_boot' and/or 'all_order' instead.").

From mathcomp Require Export all_boot.
From mathcomp Require Export preorder.
From mathcomp Require Export order.
