From mathcomp Require Import all_boot all_fingroup.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.
Local Open Scope group_scope.
Open Scope group_scope.
Check @ker_sdprodm.
